
/*
 * $Id: ObjLifeCycle.cc,v 1.2 1999/04/30 04:44:15 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include "Clock.h"
#include "Rnd.h"
#include "ObjTimes.h"
#include "ObjLifeCycle.h"

ObjLifeCycle::ObjLifeCycle():
	theModTime(0), theNowExpTime(0), theModExpTime(0),
	theModVar(0), theShowModRatio(0),
	theNowExpRatio(0), theModExpRatio(0), theShowExpRatio(0) {
}

ObjLifeCycle::~ObjLifeCycle() {
	delete theModTime;
	delete theNowExpTime;
	delete theModExpTime;
}

void ObjLifeCycle::modVar(double aVar) {
	theModVar = aVar;
}

void ObjLifeCycle::modTime(RndDistr *aDistr, double aRatio) {
	Assert(aDistr || aRatio == 0);
	delete theModTime;
	theModTime = aDistr;
	if (theModTime)
		theModTime->rndGen(&theRng);
	theShowModRatio = aRatio;
}

void ObjLifeCycle::nowExpTime(RndDistr *aDistr, double aRatio) {
	Assert(aDistr || aRatio == 0);
	delete theNowExpTime;
	theNowExpTime = aDistr;
	theNowExpRatio = aRatio;
	theShowExpRatio = theNowExpRatio + theModExpRatio;
	if (theNowExpTime)
		theNowExpTime->rndGen(&theRng);
}

void ObjLifeCycle::modExpTime(RndDistr *aDistr, double aRatio) {
	Assert(aDistr || aRatio == 0);
	delete theModExpTime;
	theModExpTime = aDistr;
	theModExpRatio = aRatio;
	theShowExpRatio = theNowExpRatio + theModExpRatio;
	if (theModExpTime)
		theModExpTime->rndGen(&theRng);
}

// calculates modification and expiration times for a given object
void ObjLifeCycle::calcTimes(int oid, ObjTimes &times) {
	// must have a mod time model to proceed
	if (!theModTime) {
		times.reset();
		return;
	}

	// seed random generator with object id
	// note: both distributions is linked to this rng
	theRng.seed(oid);

	// all time computations are in [sec]

	// several "persistent" choices
	const double expModelSeed = theRng();
	times.doShowLmt = theRng() < theShowModRatio;
	const int bday = theRng(0, 10*365*24*60*60); // born in 1970s
	const double cycle = theModTime->trial();

	// also persistent: expiration gap (used below)
	// the delta is persistent because that's how real servers
	// have it configured: mod + const_delta OR now + const_delta
	const bool knownExp = expModelSeed < theShowExpRatio;
	const double expGap = !knownExp ? 0 :
		expModelSeed < theNowExpRatio ?
			theNowExpTime->trial() :
			(cycle*theModExpTime->trial()); // convert #cycles to seconds

	// find the end of the last cycle
	const int cycleCnt = (int) floor((TheClock.time().secd()-bday) / cycle);
	const double cycleEnd = bday + cycleCnt*cycle;

	// random, cycle dependent delta
	// note the change of the seed!
	theRng.seed((int)cycleEnd);
	const double halfCycle = cycle/2;
	const double delta = theModVar * theRng(-halfCycle, halfCycle);

	// now compute the last modification time
	times.theLmt = Time::Secd(cycleEnd - halfCycle + delta);

	// calculate expiration time if needed
	if (knownExp) {
		if (expModelSeed < theNowExpRatio)
			times.theExp = TheClock + Time::Secd(expGap);
		else
			times.theExp = times.theLmt + Time::Secd(expGap);
	} else
		times.theExp = Time();
}
