
/*
 * $Id: SrvXact.cc,v 1.7 1999/04/30 23:12:24 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <strstream.h>

#include "Clock.h"
#include "Rnd.h"
#include "Connection.h"
#include "HttpDate.h"
#include "StatPing.h"
#include "StatPhaseMgr.h"
#include "SharedOpts.h"
#include "SrvOpts.h"
#include "Server.h"
#include "SrvXact.h"
#include "polyErrors.h"
#include "globals.h"


void SrvXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theReqHdr.reset();
	theTimes.reset();
	theRepHdrSize = theRepSize = -1;
	isRepCachable = false;
	theProducedSize = 0;
}

void SrvXact::exec(Server *anOwner, Connection *aConn, Time delay) {
	Assert(anOwner);
	theOwner = anOwner;
	theConn = aConn; // XXX: kludge to save connection till doStart
	if (delay > 0)
		sleepFor(delay);
	else
		doStart();
}

void SrvXact::doStart() {
	start(theConn);

	theState = stHdrWaiting;
	if (!theConn->theRd.theReserv)
		theConn->theRd.start(this);
}

void SrvXact::finish(int errNo) {
	stop(errNo);
	theOwner->noteXactDone(this);
}

void SrvXact::stopStats() {
	TheStatPhaseMgr->xactRdSz(theProducedSize, theLifeTime);
	Xaction::stopStats();
}

void SrvXact::wakeUp(const Alarm &a) {
	AlarmUser::wakeUp(a);
	doStart();
}

void SrvXact::noteReadReady(int) {
	Assert(theState == stHdrWaiting);

	theConn->read();
	if (theConn->bad()) {
		finish(-1);
		return;
	}

	// check if client decided to close our pconn
	// XXX: move this (and first read!) to conn mgr
	if (theConn->exhausted() && theConn->useCnt() > 1) {
		finish(-1); // XXX: count these (not as errors)
		return;
	}

	if (theReqHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {
		theConn->theRd.stop(this);

		// dump request headers
		if (TheOpts.doPrintReqs)
			cout.write(theConn->theRdBuf.content(), theReqHdr.theHdrSize);

		// note: clt and srv code assumes we consume only after parse!
		theConn->theRdBuf.consumed(theReqHdr.theHdrSize);

		if (theReqHdr.foreign()) {
			finish(errForeignReq); // foreign request
			return;
		}

		// note if the client will close the connection
		theConn->lastUse(!theReqHdr.keepAlive());

		makeRep(theConn->theWrBuf);

		if (!theConn->theWr.theReserv)
			theConn->theWr.start(this);
		theState = stSpaceWaiting;
		return;
	}

	if (theConn->theRdBuf.full()) { // too big header
		finish(errHugeHdr);
		return;
	}

	if (theConn->atEof()) { // premature end of headers
		finish(errPrematureEoh);
		return;
	}

	if (!theConn->theRd.theReserv) // need to read more
		theConn->theRd.start(this);
}

void SrvXact::noteWriteReady(int) {
	Assert(theState == stSpaceWaiting);
	WrBuf &buf = theConn->theWrBuf;
	Assert(theProducedSize > 0 || buf.contSize() > 0);
	
	// keep the write buffer full
	buf.pack();
	buf.appended(Min(buf.spaceSize(),
		(Size)(theRepSize - theProducedSize - buf.contSize())));

	//cerr << here << "noteWriteReady: " << theProducedSize << " < " << theRepSize << endl;

	const Size sz = theConn->write();
	if (theConn->bad()) {
		finish(-1);
		return;
	}

	Assert(sz >= 0);
	theProducedSize += sz;
	if (theProducedSize >= theRepSize) { // done
		Assert(theProducedSize == theRepSize);
		finish(0);
		return;
	}

	if (!theConn->theWr.theReserv) // need write more
		theConn->theWr.start(this);
}

// decide what kind of reply to build
void SrvXact::makeRep(WrBuf &buf) {
	// timestamps are used for both 304 and 200 replies
	TheSrvOpts.theObjLifeCycle.calcTimes(theReqHdr.theObjId, theTimes);
	// cachability is used for both 304 and 200 replies
	isRepCachable = theReqHdr.theETag & 1;

	// XXX: strstream is not the most efficient implementation
	// XXX: can be further optimized using const Strings
	const char *repStart = buf.space();
	ostrstream os(buf.space(), buf.spaceSize());

	if (theReqHdr.theIMS >= 0 && theTimes.lmt() >= 0 && 
	    theTimes.lmt() <= theReqHdr.theIMS)
		makeRep304(os);
	else
		makeRep200(os);

	buf.appended(os.tellp());

	Assert(theRepHdrSize > 0 && theRepSize >= theRepHdrSize);

	// dump reply headers
	if (TheOpts.doPrintReps)
		cout.write(repStart, theRepHdrSize);
}

// build headers for a happy 200 OK reply
void SrvXact::makeRep200(ostream &os) {
	os << "HTTP/1.0 200 OK" << crlf;

	// general-header fields:
	if (isRepCachable) {
		os << "Cache-Control: public" << crlf;
	} else {
		os << "Cache-Control: private,no-cache" << crlf;
		os << "Pragma: no-cache" << crlf;
	}

	putStdFields(os);

	if (theTimes.showLmt())
		HttpDatePrint(os << "Last-Modified: ", theTimes.lmt()) << crlf;

	const Size clen = calcRepSize();
	os << "Content-Length: " << clen << crlf;

	// extention-header fields

	os << (isRepCachable ? "X-ETag: \"11\"" : "X-ETag: \"10\"") << crlf;

	if (theReqHdr.theXactId)
		os << "X-Xact: " << theReqHdr.theXactId.genMutant() << crlf;

	os << crlf;

	theRepHdrSize = os.tellp();
	theRepSize = theRepHdrSize + clen;

	CallStat(ThePinger, chbObj(theRepSize, isRepCachable));
}

void SrvXact::makeRep304(ostream &os) {
	os << "HTTP/1.0 304 Not Modified" << crlf;
	putStdFields(os);
	os << crlf;

	theRepHdrSize = theRepSize = os.tellp();
}

// put fields acceptible for both 304 and 200 replies
void SrvXact::putStdFields(ostream &os) const {
	// general-header fields

	HttpDatePrint(os << "Date: ") << crlf;

	// keep persistent
	if (theConn->reusable())
		os << "Connection: keep-alive" << crlf;
	else
		os << "Connection: close" << crlf;

	// response-header fields (none)
	// entity-header fields

	if (theTimes.knownExp())
		HttpDatePrint(os << "Expires: ", theTimes.exp()) << crlf;
}

Size SrvXact::calcRepSize() const {
	const int objId = theReqHdr.theObjId;

	// seed random generator with objId
	TheSrvOpts.theRepSize.condDistr()->rndGen()->seed(objId);
	const Size sz = (size_t) TheSrvOpts.theRepSize();
	Assert(sz >= 0);

	// XXX: hard coded size limits:
	// most popular objects should not be too big (100*id + 10KB)
	// objects should not exceed 50MB
	static Size maxSize = Size::MB(50);
	// XXX: make these limits optional as they break const distrs
	return sz % Min((Size)(100*objId + Size::KB(10)), maxSize);
}
