
/*
 * $Id: opts.h,v 1.5 1999/04/29 22:53:06 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#ifndef OPTS_CLASSES_H
#define OPTS_CLASSES_H

#include "String.h"
#include "NetAddr.h"
#include "Rnd.h"
#include "Progress.h"
#include "Opt.h"

// bool
class BoolOpt: public Opt {
	public:
		BoolOpt(OptGrp *aGrp, const char *aName, const char *aDescr, bool def = false);

		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

		operator bool() const { return theVal; }

	protected:
		bool theVal;
};

// int
class IntOpt: public Opt {
	public:
		IntOpt(OptGrp *aGrp, const char *aName, const char *aDescr, int def = -1);

		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

		operator int() const { return theVal; }

		// not an assignment op to prevent accidents
		void set(int aVal) { theVal = aVal; }

	protected:
		int theVal;
};

// double
class DblOpt: public Opt {
	public:
		DblOpt(OptGrp *aGrp, const char *aName, const char *aDescr, double def = -1);

		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

		operator double() const { return theVal; }

	protected:
		double theVal;
};

// Time
class TimeOpt: public Opt, public Time {
	public:
		TimeOpt(OptGrp *aGrp, const char *aName, const char *aDescr, Time def = Time());

		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

	protected:
		Time &theVal;
};

// Size
class SizeOpt: public Opt, public Size {
	public:
		SizeOpt(OptGrp *aGrp, const char *aName, const char *aDescr, Size def = Size());

		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

		// not an assignment op to prevent accidents
		void set(Size aVal) { theVal = aVal; }

	protected:
		Size &theVal;
};

// String
class StrOpt: public Opt, public String {
	public:
		StrOpt(OptGrp *aGrp, const char *aName, const char *aDescr, const char *def = 0);

		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

	protected:
		String &theVal;
};

// NetAddr
class NetAddrOpt: public Opt, public NetAddr {
	public:
		NetAddrOpt(OptGrp *aGrp, const char *aName, const char *aDescr, const NetAddr &def);

		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

	protected:
		NetAddr &theVal;
};

// Rate
class RateOpt: public Opt {
	public:
		RateOpt(OptGrp *aGrp, const char *aName, const char *aDescr);

		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

		operator void *() const { return set() ? (void*)-1 : 0; }
		bool set() const { return theInterval > 0; }
		double rate() const; // [1/second]
		Time intArrTime() const; // inter-arrival time

	protected:
		double theCount;
		Time theInterval;
};

// Distr
class DistrOpt: public Opt {
	public:
		DistrOpt(OptGrp *aGrp, const char *aName, const char *aDescr, RndDistr *def = 0);

		operator void *() const { return theVal ? (void*)-1 : 0; }
		RndDistr *distr() { Assert(theVal); return theVal; }
		RndDistr *condDistr() { return theVal; }

		void argType(const String &anArgType) { theArgType = anArgType; }
		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

		double trial() { Assert(theVal); return theVal->trial(); }
		double operator ()() { return trial(); }

	protected:
		bool isRT(const String &str, double &param) const; // is right type?

	protected:
		RndDistr *theVal;
		String theArgType; // time/size/etc
};


// Help
class HelpOpt: public Opt {
	public:
		HelpOpt(OptGrp *aGrp, const char *aName, const char *aDescr);

		virtual bool parse(const char *name, const char *val);
		virtual void report() const {}
		virtual bool visible() const { return false; }
};

// Two Int
class TwoIntOpt: public Opt {
	public:
		TwoIntOpt(OptGrp *aGrp, const char *aName, const char *aDescr);

		virtual bool parse(const char *name, const char *val);
		virtual void report() const;

		bool set() const { return lo() <= hi(); }
		int hi() const { return theHiVal; }
		int lo() const { return theLoVal; }

	protected:
		int theLoVal;
		int theHiVal;
};


/* option parsing routines */

// where do these belong? parser?
extern const char *valStr(const char *val);
extern bool splitVal(const String &val, String &pref, String &suf, char del = ':');
extern bool isRatio(const String &str, double &r);
extern bool isTime(const char *str, Time &t);
extern bool isSize(const char *str, Size &s);
extern bool isDistr(const char *name, const char *val, RndDistr *&distr, const String &argType);


#endif
