
/*
 * $Id: lx.cc,v 1.2 1999/04/15 20:36:00 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include "Opt.h"
#include "opts.h"

#include "StatPing.h"
#include "StatSoft.h"
#include "StatPhase.h"
#include "ILog.h"
#include "DLog.h"
#include "LogMeta.h"
#include "ilog2dlog.h"
#include "globals.h"

bool ShutdownNow() { Assert(0); return false; }  // XXX: ARGH!

class MyOpts: public OptGrp {
	public:
		MyOpts():
			theHelpOpt(this,     "help",         "list of options"),
			theOutFileName(this, "out <file>",   "redirect console output", "-"),
			thePhaseName(this,   "phase <str>",  "phase name: warm|meas|cold", "meas"),
			theObjects(this,     "objects <str>","names of objects to extract"),
			printNames(this,     "prn_names <bool>", "prepend object values with object names")
			{}

		virtual bool validate() const;

		virtual ostream &printAnonym(ostream &os) const;
		virtual bool parseAnonym(const Array<const char *> &opts);
		virtual bool canParseAnonym() const { return true; }

	public:
		HelpOpt theHelpOpt;
		StrOpt theOutFileName;
		StrOpt thePhaseName;
		StrOpt theObjects;
		BoolOpt printNames;

		Array<String*> theFileNames;
};

class Extractor {
	public:
		Extractor(const String &aName, bool selfReg = true);
		virtual ~Extractor() {}

		const String &name() const { return theName; }
		virtual ostream &printLabel(ostream &os);

		// extracts and dumps to os
		virtual void extract(const DLog &dl, ostream &os) = 0;

		// returns a configured extractor (itself or new) or null
		virtual Extractor *configure(const String &params);

		// default configuration (to support "print all" option)
		virtual Extractor *configure() { return configure(""); }

	protected:
		String theName;
};

class HistExtractor: public Extractor {
	public:
		HistExtractor(const String &aName): Extractor(aName) {}

	protected:
		void printHist(const Histogram &h, ostream &os) const;
};


static MyOpts TheOpts;
static DLog TheDLog;

static SchArray<Extractor*> TheExtors; // all known extractors
static SchArray<Extractor*> ThePlan;   // what extractors to use

static Extractor *getExtor(const String &name);


static class RespTmPercExtr: public Extractor {
	public:
		RespTmPercExtr(): Extractor("resp_tm_perc"), thePercentile(-1) {}
		RespTmPercExtr(int aPerc);
		
		virtual ostream &printLabel(ostream &os);

		virtual void extract(const DLog &dl, ostream &os);
		virtual Extractor *configure(const String &params);
		virtual Extractor *configure() { return 0; }

	protected:
		int thePercentile; // what percentile to report
} TheRespTmPercExtr;

static struct RespTmHistExtr: public HistExtractor {
	RespTmHistExtr(): HistExtractor("resp_tm_hist") {}
	virtual void extract(const DLog &dl, ostream &os) { printHist(dl.thePhaseRec.xactLifeTime(), os); }
} TheRespTmHistExtr;

static struct RepSzHistExtr: public HistExtractor {
	RepSzHistExtr(): HistExtractor("rep_sz_hist") {}
	virtual void extract(const DLog &dl, ostream &os) { printHist(dl.thePhaseRec.theXactRdSz, os); }
} TheRepSzHistExtr;

static struct RepSzMeanExtr: public Extractor {
	RepSzMeanExtr(): Extractor("rep_sz_mean") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.thePhaseRec.theXactRdSz.stats().mean(); }
} TheRepSzMeanExtr;

static struct RespTmVsSzExtr: public Extractor {
	RespTmVsSzExtr(): Extractor("resp_tm_vs_sz") {}
	virtual void extract(const DLog &dl, ostream &os);
} TheRespTmVsSzExtr;

static struct LogCountExtr: public Extractor {
	LogCountExtr(): Extractor("log_cnt") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.theSources.count(); }
} TheLogCountExtr;

static struct ReqRateExtr: public Extractor {
	ReqRateExtr(): Extractor("req_rate") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.reqRate(); }
} TheReqRateExtr;

static struct RepRateExtr: public Extractor {
	RepRateExtr(): Extractor("rep_rate") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.repRate(); }
} TheRepRateExtr;

static struct RespTmMeanExtr: public Extractor {
	RespTmMeanExtr(): Extractor("resp_tm_mean") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.thePhaseRec.xactLifeTimeMean(); }
} TheRespTmMeanExtr;

static struct ConnTtlMeanExtr: public Extractor {
	ConnTtlMeanExtr(): Extractor("conn_ttl_mean") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.thePhaseRec.connLifeTime().stats().mean(); }
} TheConnTtlMeanExtr;

static struct ReqDHRExtr: public Extractor {
	ReqDHRExtr(): Extractor("req_dhr") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.thePhaseRec.theIdealHR.dhp(); }
} TheReqDHRExtr;

static struct RepDHRExtr: public Extractor {
	RepDHRExtr(): Extractor("rep_dhr") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.thePhaseRec.theRealHR.dhp(); }
} TheRepDHRExtr;

static struct ReqBHRExtr: public Extractor {
	ReqBHRExtr(): Extractor("req_bhr") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.thePhaseRec.theIdealHR.bhp(); }
} TheReqBHRExtr;

static struct RepBHRExtr: public Extractor {
	RepBHRExtr(): Extractor("rep_bhr") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.thePhaseRec.theRealHR.bhp(); }
} TheRepBHRExtr;

static struct ErrCntExtr: public Extractor {
	ErrCntExtr(): Extractor("err_cnt") {}
	virtual void extract(const DLog &dl, ostream &os) { os << dl.theErrCnt; }
} TheErrCntExtr;

static struct ErrPctExtr: public Extractor {
	ErrPctExtr(): Extractor("err_pct") {}
	virtual void extract(const DLog &dl, ostream &os) { os << Percent(dl.theErrCnt, dl.theErrCnt + dl.theNonErrCnt); }
} TheErrPctExtr;


/* MyOpt */

ostream &MyOpts::printAnonym(ostream &os) const {
	return os << "<log_file_name> ...";
}

bool MyOpts::parseAnonym(const Array<const char *> &opts) {
	for (int i = 0 ; i < opts.count(); ++i)
		theFileNames.append(new String(opts[i]));
	return theFileNames.count() > 0;
}

bool MyOpts::validate() const {
	// convert object names into extractors
	// XXX: misplaced, should be done in a custom Opt
	String buf = TheOpts.theObjects;
	while (buf) {
		String head, tail;
		if (splitVal(buf, head, tail, ','))
			buf = tail;
		else
			head = buf, buf = 0;

		String name, params;
		if (!splitVal(head, name, params, ':'))
			params = 0;

		if (Extractor *extor = getExtor(name)) {
			if ((extor = extor->configure(params)))
				ThePlan.append(extor);
			else
				return false;
		} else
		if (name == "all") {
			for (int i = 0; i < TheExtors.count(); ++i) {
				if (Extractor *e = TheExtors[i]->configure())
					ThePlan.append(e);
			}
		} else {
			cerr << ThePrgName << ": unknown object '" << head << "'" << endl;
			return false;
		}
	}

	if (!ThePlan.count())
		cerr << ThePrgName << ": no objects specified" << endl;
	else
		return true;
	return false;
}


/* Extractor */

Extractor::Extractor(const String &aName, bool selfReg): theName(aName) {
	if (selfReg) {
		Assert(!getExtor(aName)); // names must be unique
		TheExtors.append(this);
	}
}

ostream &Extractor::printLabel(ostream &os) {
	return os << name();
}

Extractor *Extractor::configure(const String &params) {
	if (!params || !params.len())
		return this;

	cerr << ThePrgName 
		<< ": object " << name() << " has no parameters"
		<< "; got: `" << params << "'"
		<< endl;
	return 0;
}

/* HistExtractor */

void HistExtractor::printHist(const Histogram &h, ostream &os) const {
	const int totCount = h.stats().count();
	Array<HistogramBin> percs;

	Percentiles(h, percs); // 1%
	if (percs.count() != 100) { // XXX: check when/why we have less
		os << '?';
		return;
	}

	// header
	os << endl << "#bin min max count % acc%" << endl;

	int lastIdx = -1;
	for (int i = 0; i < percs.count(); ++i) {
		const HistogramBin &b = percs[i];
		if (b.idx != lastIdx) {
			os
				<< b.idx
				<< ' ' << b.min
				<< ' ' << b.max
				<< ' ' << b.count
				<< ' ' << Percent(b.count, totCount)
				<< ' ' << Percent(b.accCount, totCount)
				<< endl;
			lastIdx = b.idx;
		}
	}
}


/* RespTmVsSzExtr */

void RespTmVsSzExtr::extract(const DLog &dl, ostream &os) {
	const int pStep = 5; // 5%;
	const TimeVsSizeHist &mh = dl.thePhaseRec.theXactTtlVsSz;

	// header
	os << endl 
		<< "#size    count response_time" << endl
		<< "#min max count mean";
	for (int i = pStep; i <= 100; i += pStep)
		os << ' ' << i << '%';
	os << endl;


	// for each size group, print mean and percentiles
	int lastGrp = mh.lastGrpId();
	int sz = 0;
	for (int g = 0, lastSz = 0; g <= lastGrp; ++g, lastSz = sz) {
		sz = mh.groupMaxSize(g);
		const Histogram &h = mh.groupHist(g);

		if (!h.stats().count())
			continue;

		// print size range, count, and mean
		os << lastSz << ' ' << (sz-1)
			<< ' ' << h.stats().count()
			<< ' ' << h.stats().mean();

		// get and print percentiles
		Array<HistogramBin> percs;
		Percentiles(h, percs);
		for (int i = pStep; i <= 100; i += pStep) {
			if (percs.count() >= i) // XXX: check when/why we have less
				os << ' ' << percs[i-1].max;
			else
				os << " ?";
		}

		os << endl;

		lastSz = sz;
	}
}


/* RespTmPercExtr */

RespTmPercExtr::RespTmPercExtr(int aPerc): 
	Extractor("resp_tm_perc", false), thePercentile(aPerc) {
	Assert(aPerc > 0 && aPerc <= 100);
}

ostream &RespTmPercExtr::printLabel(ostream &os) {
	return Extractor::printLabel(os) << ':' << thePercentile;
}

Extractor *RespTmPercExtr::configure(const String &params) {
	int perc = -1;
	const char *h;
	if (isInt(params, perc, &h) && !*h && perc > 0 && perc <= 100)
		return new RespTmPercExtr(perc);

	cerr << ThePrgName << ": integer percentile (e.g. 50 or 75) expected for"
		<< " object " << name() << "; got: `" << params << "'"
		<< endl;
	return 0;
}

void RespTmPercExtr::extract(const DLog &dl, ostream &os) {
	Assert(thePercentile > 0);
	const Histogram &h = dl.thePhaseRec.xactLifeTime();
	Array<HistogramBin> percs;
	Percentiles(h, percs); // 1%
	if (percs.count() != 100) { // XXX: check when/why we have less
		os << '?';
		return;
	}

	const HistogramBin &b = percs[thePercentile-1];
	if (h.extreme(b.idx)) // extreme bucket
		os << -1; // extreme buckets are "open-ended" and have no "average"
	else
		os << b.max; // b.max is b.max of the last bin for this pcntile
}

// extract requested objects
static
void extract(const Array<Extractor*> &plan) {

	// print all requested names unless we will do it later
	if (!TheOpts.printNames) {
		cout << "#objects:";
		for (int i = 0; i < plan.count(); ++i)
			plan[i]->printLabel(cout << ' ');
		cout << endl;
	}

	// extract actual values
	for (int i = 0; i < plan.count(); ++i) {

		if (TheOpts.printNames)
			plan[i]->printLabel(cout) << ":\t ";
		else
		if (i)
			cout << ' '; // insert delimiter

		plan[i]->extract(TheDLog, cout);

		if (TheOpts.printNames)
			cout << endl;
	}

	if (!TheOpts.printNames)
		cout << endl;
}

static 
Extractor *getExtor(const String &name) {
	for (int i = 0; i < TheExtors.count(); ++i) {
		if (TheExtors[i]->name() == name)
			return TheExtors[i];
	}
	return 0;
}

static
void configureLogs(int prec) {
	if (TheOpts.theOutFileName && strcmp(TheOpts.theOutFileName, "-"))
		cout = cerr = clog = *new ofstream(TheOpts.theOutFileName);

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

static
void configure() {
	configureLogs(2);
}

int main(int argc, char *argv[]) {

	CmdLine cmd(Array<OptGrp*>() << &TheOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate())
		return -1;

	configure();

	TheDLog.thePhaseName = TheOpts.thePhaseName;

	// suck all data in
	for (int i = 0; i < TheOpts.theFileNames.count(); ++i) {
		if (!ilog2dlog(*TheOpts.theFileNames[i], TheDLog))
			cerr << here << "error processing log " << *TheOpts.theFileNames[i] << endl;
	}

	extract(ThePlan);

	return 0;
}
